#! -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='torch4keras',
    version='0.0.1',
    description='use torch like keras',
    long_description='torch4keras: https://github.com/Tongjilibo/torch4keras',
    license='Apache License',
    url='https://github.com/Tongjilibo/torch4keras',
    author='Tongjilibo',
    install_requires=[],
    packages=find_packages()
)